/*-
 * Copyright (c) 2005 Daniel Braniss <danny@cs.huji.ac.il>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 | $Id$
 */
#include <sys/param.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "iscsi.h"
#include "iscontrol.h"
#include "pdu.h"

setOptions(isess_t *sess)
{
     isc_opt_t	oop;

     debug_called(3);

     bzero(&oop, sizeof(isc_opt_t));
     oop.initiatorName	= sess->op->initiatorName;
     oop.targetAddress	= sess->op->targetAddress;
     if(sess->op->targetName != 0)
	  oop.targetName = sess->op->targetName;

     oop.maxRecvDataSegmentLength = sess->op->maxRecvDataSegmentLength;
     oop.maxBurstLength = sess->op->maxBurstLength;
     
     if(ioctl(sess->fd, ISCSISETOPT, &oop)) {
	  perror("ISCSISETOPT");
	  return -1;
     }
     return 0;
}

static void
negotiateSome(isess_t *sess, pdu_t *spp)
{
     static int done = 0;
     isc_opt_t  *op = sess->op;

     debug_called(3);

     if(done++ == 0) {
	  // KLUDGE!!!!!!!
	  addText(spp, "HeaderDigest=%s", op->headerDigest);
	  addText(spp, "DataDigest=%s", op->dataDigest);
	  addText(spp, "MaxRecvDataSegmentLength=%d", op->maxRecvDataSegmentLength);
	  addText(spp, "MaxBurstLength=%d", op->maxBurstLength);
	  addText(spp, "FirstBurstLength=%d", op->firstBurstLength);
	  addText(spp, "DefaultTime2Wait=%d", op->defaultTime2Wait);
	  addText(spp, "ErrorRecoveryLevel=%d", op->errorRecoveryLevel);
	  addText(spp, "DefaultTime2Retain=%d", op->defaultTime2Retain);
	  addText(spp, "InitialR2T=%s", op->initialR2T? "Yes": "No");
	  addText(spp, "DataPDUInOrder=%s", op->dataPDUInOrder? "Yes": "No");
	  addText(spp, "DataSequenceInOrder=%s", op->dataSequenceInOrder? "Yes": "No");
     }
}

int
negotiateOPV(isess_t *sess)
{
     pdu_t		spp;
     login_req_t	*lp;

     debug_called(3);

     bzero(&spp, sizeof(pdu_t));
     lp = (login_req_t *)&spp.ipdu.bhs;
     lp->cmd = ISCSI_LOGIN_CMD | 0x40; // login request + Inmediate
     memcpy(lp->isid, sess->isid, 6);
     lp->tsih = sess->tsih; // MUST be zero the first time!
     lp->CID = htons(1);
     lp->CSG = LON_PHASE;	// Login Operational Negotiation
     lp->NSG = FF_PHASE;	// Full Feature
     lp->T = 1;

     negotiateSome(sess, &spp);

     return sendPDU(sess, &spp, handleLoginResp);
}

void
processParams(isess_t *sess, pdu_t *pp)
{
     isc_opt_t		*op = sess->op;
     int		len, klen, n;
     char		*eq, *ptr;

     debug_called(3);

     len = pp->ds_len;
     ptr = pp->ds;
     while(len > 0) {
	  if(vflag > 1)
	       printf("got: len=%d %s\n", len, ptr);
	  klen = 0;
	  if((eq = strchr(ptr, '=')) != NULL)
	       klen = eq - ptr;
	  if(klen > 0) {
	       if(strncmp(ptr, "TargetAddress", klen) == 0) {
		    char *p;
		    //if(op->targetAddress) free(op->targetAddress);
		    op->targetAddress = strdup(eq+1);
		    if((p = strchr(op->targetAddress, ':')) != NULL) {
			 *p++ = 0;
			 op->port = atoi(p);
		    }
	       } else
	       if(strncmp(ptr, "MaxRecvDataSegmentLength", klen) == 0) {
		    op->maxRecvDataSegmentLength = strtol(eq+1, (char **)NULL, 0);
	       } else 
	       if(strncmp(ptr, "TargetPortalGroupTag", klen) == 0) {
		    op->targetPortalGroupTag = strtol(eq+1, (char **)NULL, 0);
	       }
#if 0
	       else
	       for(kp = keyMap; kp->name; kp++) {
		    if(strncmp(ptr, kp->name, kp->len) == 0 && ptr[kp->len] == '=')
			 mp->func(sess, ptr+kp->len+1, GET);
	       }
#endif
	  }
	  n = strlen(ptr) + 1;
	  len -= n;
	  ptr += n;
     }

     setOptions(sess); // XXX: just in case ...

}
